@@ -1,3 +1,4 @@
+// ==========================================================================
 //         ____            _                     _____           _
 //        / ___| _   _ ___| |_ ___ _ __ ___     |_   _|__   ___ | |___
 //        \___ \| | | / __| __/ _ \ '_ ` _ \ _____| |/ _ \ / _ \| / __|
@@ -26,9 +27,12 @@
 //
 // Contact: thomas.dreibholz@gmail.com
 
-#define _XOPEN_SOURCE 500
+#define _XOPEN_SOURCE 700
 #include <ctype.h>
+#include <getopt.h>
 #include <fcntl.h>
+#include <libintl.h>
+#include <locale.h>
 #include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -37,6 +41,9 @@
 #include <wchar.h>
 #include <locale.h>
 #include <sys/ioctl.h>
+#ifndef nullptr
+#define nullptr NULL
+#endif
 
 #include "package-version.h"
 
@@ -55,6 +62,30 @@ typedef enum printmode {
 } printmode_t;
 
 
+static char* Utf8String  = nullptr;
+static char* BorderLeft  = nullptr;
+static char* BorderRight = nullptr;
+
+
+// ###### Clean up ##########################################################
+[[ noreturn ]] static void cleanUp(int exitCode)
+{
+   if(Utf8String) {
+      free(Utf8String);
+      Utf8String = nullptr;
+   }
+   if(BorderLeft) {
+      free(BorderLeft);
+      BorderLeft = nullptr;
+   }
+   if(BorderRight) {
+      free(BorderRight);
+      BorderRight = nullptr;
+   }
+   exit(exitCode);
+}
+
+
 // ###### Obtain console size ###############################################
 static bool ioctlTIOCGWINSZ(struct winsize* w)
 {
@@ -102,9 +133,10 @@ static unsigned int hexDigitToNumber(const char digit)
 static char* unescape(const char* originalString)
 {
    size_t original_string_length = strlen(originalString);
-   char*  unescapedString        = malloc(original_string_length + 1);
-   if(unescapedString == NULL) {
-      fputs("ERROR: malloc() failed!\n", stderr);
+   char*  unescapedString        = (char*)malloc(original_string_length + 1);
+   if(unescapedString == nullptr) {
+      fputs(gettext("ERROR: malloc() failed!"), stderr);
+      fputs("\n", stderr);
       exit(1);
    }
 
@@ -226,9 +258,10 @@ wchar_t* convertToWideStringWithoutANSI(const char* originalString,
    const size_t original_string_length = strlen(originalString);
 
    // ====== Copy string, but filter out ANSI colouring sequences ===========
-   char* utf8String = malloc(original_string_length + 1);
-   if(utf8String == NULL) {
-      fputs("ERROR: malloc() failed!\n", stderr);
+   char* utf8String = (char*)malloc(original_string_length + 1);
+   if(utf8String == nullptr) {
+      fputs(gettext("ERROR: malloc() failed!"), stderr);
+      fputs("\n", stderr);
       exit(1);
    }
 
@@ -253,13 +286,15 @@ wchar_t* convertToWideStringWithoutANSI(const char* originalString,
 
    // ====== Get string width ===============================================
    wchar_t* wide_string = (wchar_t*)malloc(sizeof(wchar_t) * (j + 1));
-   if(wide_string == NULL) {
-      fputs("ERROR: malloc() failed!\n", stderr);
+   if(wide_string == nullptr) {
+      fputs(gettext("ERROR: malloc() failed!"), stderr);
+      fputs("\n", stderr);
       exit(1);
    }
    const size_t wide_string_length = mbstowcs(wide_string, utf8String, j);
    if(wide_string_length < 0) {
-      fputs("ERROR: mbstowcs() failed!\n", stderr);
+      fputs(gettext("ERROR: mbstowcs() failed!"), stderr);
+      fputs("\n", stderr);
       exit(1);
    }
 
@@ -282,10 +317,10 @@ static int stringwidth(const char* originalString,
    size_t   wide_string_length;
    wchar_t* wide_string = convertToWideStringWithoutANSI(
                              originalString, removeANSISequences,
-                             NULL, &wide_string_length);
+                             nullptr, &wide_string_length);
    const int width = wcswidth(wide_string, wide_string_length);
    free(wide_string);
-   return width;
+   return ((width >= 0) ? width : 0);
 }
 
 
@@ -386,11 +421,18 @@ static void separator(const char* separaterBorderLeft,
                       const int   consoleWidth)
 {
    const int separaterLeftWidth   = stringwidth(separaterBorderLeft, true);
-   const int separatorStringWidth = stringwidth(separatorString, true);
    const int separaterRightWidth  = stringwidth(separaterBorderRight, true);
+   int       separatorStringWidth = stringwidth(separatorString, true);
+   int       i;
+   if(separatorStringWidth < 1) {
+      // The separator string must have a width >= 1, to fill the line.
+      // Using default " ":
+      separatorString      = " ";
+      separatorStringWidth = 1;
+   }
 
-   int i = separaterLeftWidth + separaterRightWidth;   // Border width
    fputs(separaterBorderLeft, stdout);
+   i = separaterLeftWidth + separaterRightWidth;   // Border width
    while(i + separatorStringWidth <= consoleWidth) {
       fputs(separatorString, stdout);
       i += separatorStringWidth;
@@ -414,10 +456,10 @@ static void doMultiLineIndentOrCenter(const char*       borderLeft,
    char*        s;
    char         buffer[8192];
    unsigned int maxLength = 0;
-   while( (s = fgets((char*)&buffer, sizeof(buffer), stdin)) != NULL ) {
+   while( (s = fgets((char*)&buffer, sizeof(buffer), stdin)) != nullptr ) {
       buffer[strcspn(buffer, "\r\n")] = 0x00;   // Remove newline
       lineArray[lines] = strdup(s);
-      if(lineArray[lines] == NULL) {
+      if(lineArray[lines] == nullptr) {
          exit(1);
       }
       lineLength[lines] = strlen(lineArray[lines]);
@@ -432,7 +474,7 @@ static void doMultiLineIndentOrCenter(const char*       borderLeft,
 
    // ====== Obtain left and right border widths ============================
    size_t   wll;
-   wchar_t* wsl = convertToWideStringWithoutANSI(borderLeft, true, NULL, &wll);
+   wchar_t* wsl = convertToWideStringWithoutANSI(borderLeft, true, nullptr, &wll);
    if(wsl) {
       const int borderLeftWidth = wcswidth(wsl, wll);
       if(borderLeftWidth > 0) {
@@ -441,7 +483,7 @@ static void doMultiLineIndentOrCenter(const char*       borderLeft,
       free(wsl);
    }
    size_t   wlr;
-   wchar_t* wsr = convertToWideStringWithoutANSI(borderRight, true, NULL, &wlr);
+   wchar_t* wsr = convertToWideStringWithoutANSI(borderRight, true, nullptr, &wlr);
    if(wsr) {
       const int borderRightWidth = wcswidth(wsr, wlr);
       if(borderRightWidth > 0) {
@@ -464,179 +506,261 @@ static void doMultiLineIndentOrCenter(const char*       borderLeft,
          puts("");
       }
       free(lineArray[i]);
-      lineArray[i] = NULL;
+      lineArray[i] = nullptr;
    }
 }
 
 
+// ###### Version ###########################################################
+[[ noreturn ]] static void version()
+{
+   printf("print-utf8 %s\n", SYSTEMTOOLS_VERSION);
+   exit(0);
+}
+
+
+// ###### Usage #############################################################
+[[ noreturn ]] static void usage(const char* program, const int exitCode)
+{
+   fprintf(stderr, "%s %s"
+           " [-n|--newline]"
+           " [-i indentation string|--indent indentation string]"
+           " [-c string|--center string]"
+           " [-I left right|--multiline-indent indentation left right]"
+           " [-C left right|--multiline-center left right]"
+           " [-s border_left separator border_right|--separator border_left separator border_right]"
+           " [-c columns|--columns columns]"
+           " [-s string|--size string]"
+           " [-l string|--length string]"
+           " [-w string|--width string]"
+           " [-a string|--size-length-width string]"
+           " [-t|--terminal-info]"
+           " [-h|--help]"
+           " [-v|--version]\n",
+           gettext("Usage:"),
+           program);
+   exit(exitCode);
+}
+
+
 
 // ###### Main program ######################################################
 int main (int argc, char** argv)
 {
    // ====== Initialise locale support ======================================
-   if(setlocale(LC_ALL, "") == NULL) {
+   if(setlocale(LC_ALL, "") == nullptr) {
       setlocale(LC_ALL, "C.UTF-8");   // "C" should exist on all systems!
    }
+   bindtextdomain("print-utf8", nullptr);
+   textdomain("print-utf8");
 
    // ====== Handle arguments ===============================================
    printmode_t mode                = Indent;
    int         indentWidth         = 0;
-   char*       utf8String          = NULL;
-   char*       borderLeft          = NULL;
-   char*       borderRight         = NULL;
    bool        newline             = false;
    const int   defaultConsoleWidth = getConsoleWidth();
    int         consoleWidth        = defaultConsoleWidth;
 
-   for(int i = 1; i <argc; i++) {
-      if( (strcmp(argv[i], "-i") == 0) || (strcmp(argv[i], "--indent") == 0) ) {
-         if(i + 1 < argc) {
-            indentWidth = atoi(argv[i + 1]);
-         }
-         else {
-            fputs("ERROR: Invalid indentation setting!\n", stderr);
-            return 1;
-         }
-         mode = Indent;
-         i++;
-      }
-      else if( (strcmp(argv[i], "-c") == 0) || (strcmp(argv[i], "--center") == 0) ) {
-         mode = Center;
-      }
-      else if( (strcmp(argv[i], "-I") == 0) || (strcmp(argv[i], "--multiline-indent") == 0) ) {
-         mode = MultiLineIndent;
-         if(i + 3 < argc) {
-            indentWidth = atoi(argv[i + 1]);
-            borderLeft  = unescape(argv[i + 2]);
-            borderRight = unescape(argv[i + 3]);
-         }
-         else {
-            fputs("ERROR: Invalid arguments for multiline-ident!\n", stderr);
-            return 1;
-         }
-      }
-      else if( (strcmp(argv[i], "-C") == 0) || (strcmp(argv[i], "--multiline-center") == 0) ) {
-         mode = MultiLineCenter;
-         if(i + 2 < argc) {
-            borderLeft  = unescape(argv[i + 1]);
-            borderRight = unescape(argv[i + 2]);
-         }
-         else {
-            fputs("ERROR: Invalid arguments for multiline-center!\n", stderr);
-            return 1;
-         }
-      }
-      else if( (strcmp(argv[i], "-s") == 0) || (strcmp(argv[i], "--separator") == 0) ) {
-         if(i + 3 < argc) {
-            borderLeft  = unescape(argv[i + 1]);
-            borderRight = unescape(argv[i + 3]);
-            if(utf8String) {
-               free(utf8String);
+   const static struct option long_options[] = {
+      { "indent",              required_argument, 0, 'i' },
+      { "multiline-indent",    required_argument, 0, 'I' },
+      { "center",              no_argument,       0, 'c' },
+      { "multiline-center",    required_argument, 0, 'C' },
+      { "separator",           required_argument, 0, 's' },
+
+      { "newline",             no_argument,       0, 'n' },
+      { "terminal-info",       no_argument,       0, 't' },
+      { "width",               no_argument,       0, 'w' },
+      { "length",              no_argument,       0, 'l' },
+      { "size",                no_argument,       0, 'b' },
+      { "size-length-width",   no_argument,       0, 'a' },
+      { "columns",             required_argument, 0, 'x' },
+
+      { "",                    no_argument,       0, '-' },
+      { "help",                no_argument,       0, 'h' },
+      { "version",             no_argument,       0, 'v' },
+      {  nullptr,              0,                 0, 0   }
+   };
+
+   int option;
+   int longIndex;
+   while( (option = getopt_long(argc, argv, "i:cI:C:s:ntwlbax:-hv", long_options, &longIndex)) != -1 ) {
+      // NOTE: optind already points to the next option here!
+      //       For options with two or more parameters, this
+      //       will be the *second* parameter!
+      switch(option) {
+         case 'i':
+            mode        = Indent;
+            indentWidth = atoi(optarg);
+          break;
+         case 'c':
+            mode = Center;
+          break;
+         case 'I':
+            mode = MultiLineIndent;
+            if(optind + 1 < argc) {
+               indentWidth = atoi(argv[optind - 1]);
+               if(BorderLeft) {
+                  free(BorderLeft);
+               }
+               BorderLeft  = unescape(argv[optind + 0]);
+               if(BorderRight) {
+                  free(BorderRight);
+               }
+               BorderRight = unescape(argv[optind + 1]);
+               optind += 2;
             }
-            utf8String  = unescape(argv[i + 2]);
-         }
-         else {
-            fputs("ERROR: Invalid separator setting!\n", stderr);
-            return 1;
-         }
-         mode = Separator;
-         i += 3;
-      }
-      else if( (strcmp(argv[i], "-n") == 0) || (strcmp(argv[i], "--newline") == 0) ) {
-         newline = true;
-      }
-      else if( (strcmp(argv[i], "-t") == 0) || (strcmp(argv[i], "--terminal-info") == 0) ) {
-         mode = TerminalInfo;
-      }
-      else if( (strcmp(argv[i], "-x") == 0) || (strcmp(argv[i], "--columns") == 0) ) {
-         if(i + 1 < argc) {
-            consoleWidth = atol(argv[i + 1]);
-            if( consoleWidth <= 0) {
-               consoleWidth = defaultConsoleWidth + consoleWidth;   // subtract!
+            else {
+               fputs(gettext("ERROR: Invalid arguments for multiline-indent!"), stderr);
+               fputs("\n", stderr);
+               cleanUp(1);
             }
-            if (consoleWidth > 4096) {
-               fprintf(stderr, "ERROR: Invalid console width %u!\n", consoleWidth);
-               return 1;
+          break;
+         case 'C':
+            mode = MultiLineCenter;
+            if(optind < argc) {
+               if(BorderLeft) {
+                  free(BorderLeft);
+               }
+               BorderLeft  = unescape(argv[optind - 1]);
+               if(BorderRight) {
+                  free(BorderRight);
+               }
+               BorderRight = unescape(argv[optind - 0]);
+               optind++;
             }
-         }
-         i += 1;
+            else {
+               fputs(gettext("ERROR: Invalid arguments for multiline-center!"), stderr);
+               fputs("\n", stderr);
+               cleanUp(1);
+            }
+          break;
+         case 's':
+            mode = Separator;
+            if(optind + 1 < argc) {
+               if(BorderLeft) {
+                  free(BorderLeft);
+               }
+               BorderLeft  = unescape(argv[optind - 1]);
+               if(BorderRight) {
+                  free(BorderRight);
+               }
+               BorderRight = unescape(argv[optind + 1]);
+               if(Utf8String) {
+                  free(Utf8String);
+               }
+               Utf8String  = unescape(argv[optind + 0]);
+               optind += 2;
+            }
+            else {
+               fputs(gettext("ERROR: Invalid separator setting!"), stderr);
+               fputs("\n", stderr);
+               cleanUp(1);
+            }
+          break;
+         case 'x':
+            if(optind < argc) {
+               consoleWidth = atol(optarg);
+               if(consoleWidth <= 0) {
+                  consoleWidth = defaultConsoleWidth + consoleWidth;   // subtract!
+               }
+               if(consoleWidth > 4096) {
+                  fprintf(stderr, gettext("ERROR: Invalid console width %u!"), consoleWidth);
+                  fputs("\n", stderr);
+                  cleanUp(1);
+               }
+            }
+          break;
+         case 'n':
+            newline = true;
+          break;
+         case 't':
+            mode = TerminalInfo;
+          break;
+         case 'w':
+            mode = Width;
+          break;
+         case 'l':
+            mode = Length;
+          break;
+         case 'b':
+            mode = Size;
+          break;
+         case 'a':
+            mode = SizeLengthWidth;
+          break;
+         case 'v':
+            version();
+          break;
+         case 'h':
+            usage(argv[0], 0);
+          break;
+         case '-':
+          break;
+         default:
+          break;
       }
-      else if( (strcmp(argv[i], "-w") == 0) || (strcmp(argv[i], "--width") == 0) ) {
-         mode = Width;
+   }
+   if(optind < argc) {
+      if(Utf8String) {
+         free(Utf8String);
       }
-      else if( (strcmp(argv[i], "-l") == 0) || (strcmp(argv[i], "--length") == 0) ) {
-         mode = Length;
-      }
-      else if( (strcmp(argv[i], "-b") == 0) || (strcmp(argv[i], "--size") == 0) ) {
-         mode = Size;
-      }
-      else if( (strcmp(argv[i], "-a") == 0) || (strcmp(argv[i], "--size-length-width") == 0) ) {
-         mode = SizeLengthWidth;
-      }
-      else if( (strcmp(argv[i], "-h") == 0) || (strcmp(argv[i], "--help") == 0) ) {
-         fprintf(stderr, "Usage: %s [-n|--newline] [-i indentation string|--ident indentation string] [-c string|--center string] [-I left right|--multiline-indent indentation left right]  [-C left right|--multiline-center left right] [-s border_left separator border_right|--separator border_left separator border_right] [-c columns|--columns columns] [-s string|--size string] [-l string|--length string] [-w string|--width string] [-a string|--size-length-width string] [-t|--terminal-info] [-h|--help] [-v|--version]\n", argv[0]);
-         return 1;
-      }
-      else if( (strcmp(argv[i], "-v") == 0) || (strcmp(argv[i], "--version") == 0) ) {
-         printf("printf-utf8 %s\n", SYSTEMTOOLS_VERSION);
-         return 0;
-      }
-      else {
-         if(utf8String) {
-            free(utf8String);
-         }
-         utf8String = unescape(argv[i]);
-         break;
+      Utf8String = unescape(argv[optind++]);
+      while(optind < argc) {
+         fprintf(stderr, gettext("WARNING: Unhandled parameter %s!"), argv[optind++]);
+         fputs("\n", stderr);
       }
    }
 
    // ====== Handle command =================================================
-   if( (utf8String) || (mode ==  TerminalInfo) ) {
-      switch(mode) {
-         case Indent:
-            indented(utf8String, indentWidth);
-          break;
-         case Center:
-            centered(utf8String, consoleWidth, false);
-          break;
-         case MultiLineIndent:
-         case MultiLineCenter:
-            doMultiLineIndentOrCenter(borderLeft, borderRight,
-                                      consoleWidth, indentWidth, mode);
-          break;
-         case Separator:
-            separator(borderLeft, utf8String, borderRight, consoleWidth);
-          break;
-         case Width:
-            stringSizeLengthWidth(utf8String, true, false, false, true);
-          break;
-         case Length:
-            stringSizeLengthWidth(utf8String, true, false, true, false);
-          break;
-         case Size:
-            stringSizeLengthWidth(utf8String, true, true, false, false);
-          break;
-         case SizeLengthWidth:
-            stringSizeLengthWidth(utf8String, true, true, true, true);
-          break;
-         case TerminalInfo:
-            terminalInfo();
-          break;
-      }
+   switch(mode) {
+      case Indent:
+         if(Utf8String) {
+            indented(Utf8String, indentWidth);
+         }
+         break;
+      case Center:
+         if(Utf8String) {
+            centered(Utf8String, consoleWidth, false);
+         }
+         break;
+      case MultiLineIndent:
+      case MultiLineCenter:
+         doMultiLineIndentOrCenter(BorderLeft, BorderRight,
+                                    consoleWidth, indentWidth, mode);
+         break;
+      case Separator:
+         if(Utf8String) {
+            separator(BorderLeft, Utf8String, BorderRight, consoleWidth);
+         }
+         break;
+      case Width:
+         if(Utf8String) {
+            stringSizeLengthWidth(Utf8String, true, false, false, true);
+         }
+         break;
+      case Length:
+         if(Utf8String) {
+            stringSizeLengthWidth(Utf8String, true, false, true, false);
+         }
+         break;
+      case Size:
+         if(Utf8String) {
+            stringSizeLengthWidth(Utf8String, true, true, false, false);
+         }
+         break;
+      case SizeLengthWidth:
+         if(Utf8String) {
+            stringSizeLengthWidth(Utf8String, true, true, true, true);
+         }
+         break;
+      case TerminalInfo:
+         terminalInfo();
+         break;
    }
    if(newline) {
       fputs("\n", stdout);
    }
 
-   if(utf8String) {
-      free(utf8String);
-   }
-   if(borderLeft) {
-      free(borderLeft);
-   }
-   if(borderRight) {
-      free(borderRight);
-   }
-
-   return 0;
+   cleanUp(0);
 }
